//  //
//  ###  ##### #   #               -= Confix Library =-                  //
// #   # #     ## ## ConfigSection.h - Configuration file section        //
// #     ####   ###                                                      //
// #     #      ###  Representation of a configuration file section      //
// #   # #     ## ##                                                     //
//  ###  #     #   # R1             (C)2002 Markus Ewald -> License.txt  //
//  //
#ifndef CFX_CONFIGSECTION_H
#define CFX_CONFIGSECTION_H

#include "Confix/Config.h"
#include "Confix/ConfigVariable.h"

#include <string>
#include <vector>

#ifdef CONFIX_NAMESPACE
  namespace CONFIX_NAMESPACE {
#endif CONFIX_NAMESPACE

//  //
//  CFX::CConfigSection                                                  //
//  //
/// A configuration file section
/** Stores the contents of a section in the configuration file. Sections
    are declared in brackets (eg. '[Section]') and enable the grouping of
    values to further organize the configuration file.

    All variables declared in the configuration file before the first
    section begins are added to a nameless section.
*/
class CONFIX_API CConfigSection {
  public:
    /// Config variable enumerator
    class CONFIX_API CVariableEnumerator;

    /// Constructor
    CConfigSection(void);
    /// Constructor
    CConfigSection(const std::string &sContents);
    /// Copy constructor
    CConfigSection(const CConfigSection &Section);
    /// Destructor
    ~CConfigSection(void);

  //
  // CConfigSection implementation
  //
  public:
    /// Get section name
    const std::string &getName(void) const;
    /// Set section name
    void setName(const std::string &sName);

    /// Get number of variables in section
    unsigned long getNumVariables(void) const;

    /// Get variable
    CConfigVariable &getVariable(unsigned long lIndex);
    /// Get variable 
    CConfigVariable &getVariable(const std::string &sName);

    /// Add variable
    void addVariable(const CConfigVariable &Variable);
    /// Remove variable
    void removeVariable(unsigned long lIndex);
    /// Remove variable
    void removeVariable(const std::string &sName);

    /// Get variable enumerator
    CVariableEnumerator getVariables(void);

    /// Clear section
    void clearContents(void);
    /// Get configuration contents
    std::string getContents(void) const;
    /// Set configuration contents
    std::string::size_type setContents(const std::string &sContents, std::string::size_type Start = 0);

  private:
    /// A map of configuration variables
    typedef std::vector<CConfigVariable> VariableVector;

    std::string    *m_psName;                         ///< Section name
    VariableVector *m_pVariables;                     ///< Variables in section

    mutable bool                    m_bChanged;       ///< Has changed
    mutable std::string            *m_psContents;     ///< Section line
    mutable std::string::size_type  m_NameStart;      ///< Name starting offset
    mutable std::string::size_type  m_NameLength;     ///< Name length
};

/** Enumerates over a list of config variables. An enumerator is a simple
    forward iterator which allows you to wander through a list of items
    without knowing the underlying data structure used for their storage.

    The enumerator starts in an empty position, meaning the first method
    you have to call is toNext(). This method returns true as long as it
    could advance to the next item. If there are no items available to be
    enumerated at all, toNext() will immediately return false.

    Usage is simple:
    @code
    CConfigSection::CVariableEnumerator VarEnum = MySection.getVariables();
    while(VarEnum.toNext())
      cout << VarEnum.getVariable().getName() << endl;
    @endcode
*/
class CONFIX_API CConfigSection::CVariableEnumerator {
  public:
    /// Constructor
    CVariableEnumerator(VariableVector &Variables);
    /// Copy constructor
    CVariableEnumerator(const CVariableEnumerator &VariableEnum);
    /// Destructor
    ~CVariableEnumerator(void);

  //
  // CVariableEnumerator implementation
  //
  public:
    /// Get current variable
    CConfigVariable &getVariable(void);

    /// Jump to next variable
    bool toNext(void);

  private:
    VariableVector           &m_Variables;            ///< Variable vector
    VariableVector::iterator  m_VariableIt;           ///< Current variable
    VariableVector::iterator  m_VariableEnd;          ///< End variable
};

#ifdef CONFIX_NAMESPACE
  }; // CONFIX_NAMESPACE
#endif CONFIX_NAMESPACE

#endif CFX_CONFIGVARIABLE_H
