//  //
//  ###  ##### #   #               -= Confix Library =-                  //
// #   # #     ## ## ConfigSection.cpp - Configuration file section      //
// #     ####   ###                                                      //
// #     #      ###  Representation of a configuration file section      //
// #   # #     ## ##                                                     //
//  ###  #     #   # R1             (C)2002 Markus Ewald -> License.txt  //
//  //
#include "Confix/ConfigSection.h"
#include <stdexcept>
#include <stdio.h> //!!TEMP
#include "../../Demo/TimeMonitor.h"

#ifdef CONFIX_NAMESPACE
  using namespace CONFIX_NAMESPACE;
#endif CONFIX_NAMESPACE

// ####################################################################### //
// # CFX::CConfigSection::CConfigSection()                   Constructor # //
// ####################################################################### //
/** Initializes an instance of CConfigSection
*/
CConfigSection::CConfigSection(void) :
  m_psName(new std::string()),
  m_pVariables(new VariableVector()),
  m_bChanged(false),
  m_psContents(NULL) {
  return;
}

// ####################################################################### //
// # CFX::CConfigSection::CConfigSection()                   Constructor # //
// ####################################################################### //
/** Initializes an instance of CConfigSection

    @param  sContents  Initial contents string
*/
CConfigSection::CConfigSection(const std::string &sContents) :
  m_psName(new std::string()),
  m_pVariables(new VariableVector()),
  m_bChanged(false),
  m_psContents(NULL) {

  if(sContents[sContents.length() - 1] != '\n')
    setContents(sContents + '\n');
  else
    setContents(sContents);

  return;
}


// ####################################################################### //
// # CFX::CConfigSection::CConfigSection()                   Constructor # //
// ####################################################################### //
/** Initializes an instance of CConfigSection as copy of an
    existing instance
    
    @param  Section  Section to copy
*/
CConfigSection::CConfigSection(const CConfigSection &Section) :
  m_psName(new std::string(*Section.m_psName)),
  m_pVariables(new VariableVector(*Section.m_pVariables)),
  m_bChanged(Section.m_bChanged),
  m_psContents(NULL),
  m_NameStart(Section.m_NameStart),
  m_NameLength(Section.m_NameLength) {

  if(Section.m_psContents)
    m_psContents = new std::string(*Section.m_psContents);

  return;
}

// ####################################################################### //
// # CFX::CConfigSection::~CConfigSection()                   Destructor # //
// ####################################################################### //
/** Destroys an instance of CConfigSection
*/
CConfigSection::~CConfigSection(void) {
  clearContents();
  delete m_pVariables;
  delete m_psName;

  return;
}

// ####################################################################### //
// # CFX::CConfigSection::clearContents()                                # //
// ####################################################################### //
/** Clears the section's contents
*/
void CConfigSection::clearContents(void) {
  delete m_psContents;
  m_psContents = NULL;

  *m_psName = "";
  m_pVariables->clear();

  return;
}

// ####################################################################### //
// # CFX::CConfigSection::getContents()                                  # //
// ####################################################################### //
/** Retrieves the configuration file contents for this variable

    @return The configuration file contents for the variable
*/
std::string CConfigSection::getContents(void) const { 
  std::string sContents;

  if(!m_bChanged) {
    if(m_psContents)
      sContents = *m_psContents;
    else
      sContents = "";
  } else {
    if(m_psContents) {
      *m_psContents = m_psContents->substr(0, m_NameStart) + *m_psName + m_psContents->substr(m_NameStart + m_NameLength);
      m_NameLength = m_psName->length();
      m_bChanged = false;

      sContents = *m_psContents;
    } else {
      m_psContents = new std::string(std::string("[") + *m_psName + "]\n");
      m_NameStart = 1;
      m_NameLength = m_psName->length();
      m_bChanged = false;

      sContents = *m_psContents;
    }
  }

  for(unsigned long lVariable = 0; lVariable < m_pVariables->size(); lVariable++)
    sContents += m_pVariables->operator[](lVariable).getContents();

  return sContents;
}

// ####################################################################### //
// # CFX::CConfigSection::setContents()                                  # //
// ####################################################################### //
/** Sets the configuration file contents for this variable

    @param  sContents  The configuration file contents for the variable
*/
std::string::size_type CConfigSection::setContents(const std::string &sContents, std::string::size_type Start) {
  std::string::size_type Pos = Start;

  clearContents();

  // Ensure that the string is not empty
  if(sContents.length() != 0) {
    // Look for a section declaration
    Pos = sContents.find_first_not_of(" \t", Start);
    if((Pos != std::string::npos) && (sContents[Pos] == '[')) {
      Pos++;

      // Look for the first non-spacing character in the section
      while(Pos < sContents.length()) {
        std::string::value_type Char = sContents[Pos];

        if((Char != ' ') && (Char != '\t'))
          break;

        Pos++;
      }
      std::string::size_type LastNonSpacePos = Pos;

      // Scan the remaining string for a closing bracket
      m_NameStart = Pos;
      while(Pos < sContents.length()) {
        std::string::value_type Char = sContents[Pos];
        
        if((Char == ']') || (Char == '\n'))
          break;
        
        if(Char == '\r') {
          if(sContents[Pos + 1] == '\n') {
            Pos++;
            break;
          }
        } else if((Char != ' ') && (Char != '\t')) {
          LastNonSpacePos = Pos + 1;
        }

        Pos++;
      }
      Pos++;

      m_NameLength = LastNonSpacePos - m_NameStart;

      // The section name was extracted successfully
      *m_psName = sContents.substr(m_NameStart, m_NameLength);

      Pos = sContents.find_first_of('\n', Pos) + 1;
      m_NameStart -= Start;
      m_psContents = new std::string(sContents.substr(Start, Pos - Start));

    } else {
      Pos = Start;
      m_psContents = NULL; //new std::string();
    }

    // Parse all variables in this section
    while(Pos < sContents.length()) { 
      std::string::value_type Char;

      std::string::size_type Bracket = Pos;
      while(Bracket < sContents.length()) {
        Char = sContents[Bracket];

        if((Char != ' ') && (Char != '\t'))
          break;

        Bracket++;
      }

      // Stop parsing when a new section begins
      if(Char == '[') {
        break;

      } else {
        CConfigVariable Variable;

        Pos = Variable.setContents(sContents, Pos);

        if(m_pVariables->size() >= m_pVariables->capacity())
          m_pVariables->reserve(m_pVariables->size() + 16);

        m_pVariables->push_back(Variable);
      }
    }
    m_pVariables->reserve(m_pVariables->size());

  } else {
    m_psContents = new std::string();
  }

  return Pos;
}

// ####################################################################### //
// # CFX::CConfigSection::getName()                                      # //
// ####################################################################### //
/** Retrieves the name of the configuration section

    @return The section's name
*/
const std::string &CConfigSection::getName(void) const {
  return *m_psName;
}

// ####################################################################### //
// # CFX::CConfigSection::setName()                                      # //
// ####################################################################### //
/** Sets the name of the configuration section

    @param  sName  New name for the section
*/
void CConfigSection::setName(const std::string &sName) {
  *m_psName = sName;
  m_bChanged = true;
  return;
}

// ####################################################################### //
// # CFX::CConfigSection::getNumVariables()                              # //
// ####################################################################### //
/** Get number of variables in the config section

    @return The total number of variables
*/
unsigned long CConfigSection::getNumVariables(void) const {
  return m_pVariables->size();
}

// ####################################################################### //
// # CFX::CConfigSection::getVariable()                                  # //
// ####################################################################### //
/** Retrieve a variable from the section

    @param  lIndex  Index of the variable to retrieve
    @return The variable's index
*/
CConfigVariable &CConfigSection::getVariable(unsigned long lIndex) {
  if(lIndex >= m_pVariables->size())
    throw std::out_of_range("The specified index is out of range");

  return m_pVariables->operator[](lIndex);
}

// ####################################################################### //
// # CFX::CConfigSection::getVariable()                                  # //
// ####################################################################### //
/** Retrieve a variable from the section

    @param  sName  Name of the variable to retrieve
    @return The variable's index
*/
CConfigVariable &CConfigSection::getVariable(const std::string &sName) {
  VariableVector::iterator VariableEnd = m_pVariables->end();
  for(VariableVector::iterator VariableIt = m_pVariables->begin();
      VariableIt != VariableEnd;
      VariableIt++) {
    if(VariableIt->getName() == sName)
      return *VariableIt;
  }

  throw std::invalid_argument("A variable with the specified name could not be found");
}

// ####################################################################### //
// # CFX::CConfigSection::addVariable()                                  # //
// ####################################################################### //
/** Add a variable to the section

    @param  pVariable  Variable to add
    @return The variable's index
*/
void CConfigSection::addVariable(const CConfigVariable &Variable) {
  m_pVariables->push_back(Variable);

  return;
}

// ####################################################################### //
// # CFX::CConfigSection::removeVariable()                               # //
// ####################################################################### //
/** Remove a variable from the section

    @param  lIndex  Index of the variable to remove
*/
void CConfigSection::removeVariable(unsigned long lIndex) {
  if(lIndex >= m_pVariables->size())
    throw std::out_of_range("The specified index is out of range");

  m_pVariables->erase(m_pVariables->begin() + lIndex);

  return;
}

// ####################################################################### //
// # CFX::CConfigSection::removeVariable()                               # //
// ####################################################################### //
/** Remove a variable from the section

    @param  sName  Name of the variable to remove
*/
void CConfigSection::removeVariable(const std::string &sName) {
  bool bFound = false;

  VariableVector::iterator VariableEnd = m_pVariables->end();
  for(VariableVector::iterator VariableIt = m_pVariables->begin();
      VariableIt != VariableEnd;
      VariableIt++) {
    if(VariableIt->getName() == sName) {
      m_pVariables->erase(VariableIt);
      bFound = true;
    }
  }

  if(!bFound)
    throw std::invalid_argument("A variable with the specified name could not be found");
}

// ####################################################################### //
// # CFX::CConfigSection::getVariables()                                 # //
// ####################################################################### //
/** Get a new variable enumerator

    @return A new variable enumerator
*/
CConfigSection::CVariableEnumerator CConfigSection::getVariables(void) {
  return CVariableEnumerator(*m_pVariables);
}

// ####################################################################### //
// # CFX::CConfigSection::CVariableEnumerator::CVariableEnumerator()     # Constructor //
// ####################################################################### //
/** Initializes an instance of CVariableEnumerator 

    @param  Variables  The variables which to enumerate
*/
CConfigSection::CVariableEnumerator::CVariableEnumerator(VariableVector &Variables) :
  m_Variables(Variables),
  m_VariableIt(Variables.end()),
  m_VariableEnd(m_VariableIt) {

  return;
}

// ####################################################################### //
// # CFX::CConfigSection::CVariableEnumerator::~CVariableEnumerator()    # Destructor //
// ####################################################################### //
/** Destroys an instance of CVariableEnumerator 
*/
CConfigSection::CVariableEnumerator::~CVariableEnumerator(void) {
  return;
}

// ####################################################################### //
// # CFX::CConfigSection::CVariableEnumerator::getVariable()             # //
// ####################################################################### //
/** Returns the current config variable

    @return The current config variable
*/
CConfigVariable &CConfigSection::CVariableEnumerator::getVariable(void) {
  if(m_VariableIt == m_VariableEnd)
    throw std::out_of_range("The enumerator is in the empty cycle position");

  return *m_VariableIt;
}

// ####################################################################### //
// # CFX::CConfigSection::CVariableEnumerator::toNext()                  # //
// ####################################################################### //
/** Advances to the next variable        

    @return True until the end is reached
*/
bool CConfigSection::CVariableEnumerator::toNext(void) { 
  if(m_VariableIt == m_VariableEnd)
    m_VariableIt = m_Variables.begin();
  else
    m_VariableIt++;

  if(m_VariableIt != m_VariableEnd)
    return true;
  else
    return false;
}
